/*
 * apple2.c
 *
 * driver for Apple ][ computers
 */

/* $Id: apple2.c,v 1.2 2000/01/01 20:54:23 nyef Exp $ */

#include <stdlib.h>

#include "ui.h"
#include "cal.h"
#include "types.h"
#include "tool.h"
#include "apple2_vdp.h"

struct apple2_mainboard {
    cal_cpu cpu;
    apple2_vdp vdp;
    u8 *memory;

    u8 *rom_f8;
} apple2_true;

struct apple2_mainboard *apple2 = &apple2_true;

u8 apple2_read_io(u32 addr)
{
/*      deb_printf("io read 0x%04x.\n", addr); */
    if (addr < 0xc080) { /* Built-In I/O */
	if (addr < 0xc010) { /* keyboard */
	    return 0x00; /* FIXME: add keyboard */
	} else if (addr < 0xc020) { /* keyboard strobe */
	    return 0x00; /* FIXME: add keyboard */
	} else if (addr < 0xc050) { /* unimplemented */
	    return 0x00;
	} else if (addr < 0xc058) { /* Video */
	    a2v_io(apple2->vdp, addr);
	    return 0x00;
	}
    }
    return 0x00;
}

void apple2_write_io(u32 addr, u8 data)
{
/*      deb_printf("io write 0x%04x, 0x%02x.\n", addr, data); */
    if (addr < 0xc080) { /* Built-In I/O */
	if (addr < 0xc010) { /* keyboard, read-only */
	} else if (addr < 0xc020) { /* keyboard strobe */
	    /* FIXME: add keyboard */
	} else if (addr < 0xc050) { /* unimplemented */
	    return;
	} else if (addr < 0xc058) { /* Video */
	    a2v_io(apple2->vdp, addr);
	    return;
	}
    }
}

u8 apple2_read_ram(cal_cpu cpu, u32 addr)
{
    return apple2->memory[addr];
}

void apple2_write_ram(cal_cpu cpu, u32 addr, u8 data)
{
    apple2->memory[addr] = data;
}

u8 apple2_read_rom(cal_cpu cpu, u32 addr)
{
    if (addr < 0xd000) {
	return apple2_read_io(addr);
    } else if (addr >= 0xf800) {
	return apple2->rom_f8[addr - 0xf800];
    } else {
	return 0x00;
    }
}

void apple2_write_rom(cal_cpu cpu, u32 addr, u8 data)
{
    if (addr < 0xd000) {
	apple2_write_io(addr, data);
    }
}

memread8_t apple2_readtable[4] = {
    apple2_read_ram,
    apple2_read_ram,
    apple2_read_ram,
    apple2_read_rom,
};

memwrite8_t apple2_writetable[4] = {
    apple2_write_ram,
    apple2_write_ram,
    apple2_write_ram,
    apple2_write_rom,
};

void apple2_timeslice(struct apple2_mainboard *apple2)
{
    int i;

    for (i = 243; i && (system_flags == F_NONE); i--) {
	if (system_flags == F_NONE) {
	    apple2->cpu->runfor(apple2->cpu, 65);
	    apple2->cpu->run(apple2->cpu);
	    a2v_periodic(apple2->vdp);
	    if (system_flags & F_UNIMPL) {
		deb_printf("Unimplimented Instruction\n");
		system_flags |= F_QUIT;
	    }
	}
    }
}

void apple2_load_bios(void)
{
    rom_file romfile;
    
    deb_printf("attempting to load F8 ROM.\n");
    
    romfile = read_romimage("monitor_f8.rom");
/*      romfile = read_romimage("autostart_f8.rom"); */

    if (!romfile) {
	deb_printf("could not open F8 ROM.\n");
	deb_printf("things are about to get ugly.\n");
	return;
    }
    
    apple2->rom_f8 = romfile->data;

    /* FIXME: memory leak */
}

void apple2_init_cpu(void)
{
    cal_cpu cpu;
    
    cpu = cal_create(CPUT_MARAT6502);
    apple2->cpu = cpu;
    cpu->setzpage(cpu, apple2->memory);
    cpu->setmmu8(cpu, 14, 3, apple2_readtable, apple2_writetable);
    cpu->reset(cpu);
}

void apple2_run(rom_file romfile)
{
    if (romfile) {
	deb_printf("apple2_run: ignoring romfile.\n");
    }

    apple2->memory = malloc(16384 * 3);

    if (!apple2->memory) {
	deb_printf("unable to allocate system memory.\n");
	return;
    }
    
    apple2_load_bios();

    apple2_init_cpu();
    
    apple2->vdp = a2v_init(apple2->memory); /* FIXME: return value? */

    set_timeslice((void (*)(void *)) &apple2_timeslice, apple2);
}

/*
 * $Log: apple2.c,v $
 * Revision 1.2  2000/01/01 20:54:23  nyef
 * fixed to use the correct number of cycles per line
 *
 * Revision 1.1  2000/01/01 04:12:14  nyef
 * Initial revision
 *
 */
