/*
 * apple2_vdp.c
 *
 * emulation of the Apple ][ video display
 */

/* $Id: apple2_vdp.c,v 1.1 2000/01/01 04:12:20 nyef Exp $ */

#include <stdlib.h>
#include <string.h>

#include "video.h"
#include "ui.h"
#include "tool.h"
#include "apple2_vdp.h"

int a2v_palbase_red[8] = {
    0, 0, 230, 255, 0, 230, 0, 255,
};

int a2v_palbase_green[8] = {
    0, 255, 0, 255, 0, 230, 0, 255,
};

int a2v_palbase_blue[8] = {
    0, 0, 230, 255, 0, 0, 255, 255,
};

struct apple2_vdp {
    u8 *memory;
    u8 *chargen;
    u16 scanline;
    int is_textmode;
    int is_hires;
    int is_mixed;
    int is_primary;
};

void a2v_io(apple2_vdp vdp, int addr)
{
    deb_printf("a2v_io: switch %d.\n", addr & 7);
    switch (addr & 7) {
    case 0: /* $c050 */
	vdp->is_textmode = 0;
	break;

    case 1: /* $c051 */
	vdp->is_textmode = 1;
	break;

    case 2: /* $c052 */
	vdp->is_mixed = 0;
	break;

    case 3: /* $c053 */
	vdp->is_mixed = 1;
	break;

    case 4: /* $c054 */
	vdp->is_primary = 1;
	break;

    case 5: /* $c055 */
	vdp->is_primary = 0;
	break;

    case 6: /* $c056 */
	vdp->is_hires = 0;
	break;

    case 7: /* $c057 */
	vdp-> is_hires = 1;
	break;
    }
}

int a2v_text_membase[24] = {
    0x000, 0x080, 0x100, 0x180, 0x200, 0x280, 0x300, 0x380,
    0x028, 0x0a8, 0x128, 0x1a8, 0x228, 0x2a8, 0x328, 0x3a8,
    0x050, 0x0d0, 0x150, 0x1d0, 0x250, 0x2d0, 0x350, 0x3d0,
};

void a2v_render_text(apple2_vdp vdp)
{
    u8 *cur_vbp;
    u8 *membase;
    int i;
    u8 chardata;
    
    cur_vbp = video_get_vbp(vdp->scanline);

    membase = vdp->memory;
    
    if (vdp->is_primary) {
	membase += 0x400;
    } else {
	membase += 0x800;
    }

    membase += a2v_text_membase[vdp->scanline >> 3];
    
    for (i = 0; i < 40; i++) {
	chardata = vdp->chargen[(membase[i] << 3) + (vdp->scanline & 7)];
	*cur_vbp++ = vid_pre_xlat[(chardata >> 1) & 1];
	*cur_vbp++ = vid_pre_xlat[(chardata >> 2) & 1];
	*cur_vbp++ = vid_pre_xlat[(chardata >> 3) & 1];
	*cur_vbp++ = vid_pre_xlat[(chardata >> 4) & 1];
	*cur_vbp++ = vid_pre_xlat[(chardata >> 5) & 1];
	*cur_vbp++ = vid_pre_xlat[(chardata >> 6) & 1];
	*cur_vbp++ = vid_pre_xlat[(chardata >> 7) & 1];
    }
}

void a2v_render_hires(apple2_vdp vdp)
{
    u8 *cur_vbp;

    cur_vbp = video_get_vbp(vdp->scanline);

    memset(cur_vbp, vid_pre_xlat[2], 280);
}

void a2v_render_lores(apple2_vdp vdp)
{
    u8 *cur_vbp;

    cur_vbp = video_get_vbp(vdp->scanline);

    memset(cur_vbp, vid_pre_xlat[5], 280);
}

void a2v_periodic(apple2_vdp vdp)
{
    if (vdp->scanline < 192) {
	if (vdp->is_textmode) {
	    a2v_render_text(vdp);
	} else if ((vdp->scanline >= 160) && (vdp->is_mixed)) {
	    a2v_render_text(vdp);
	} else if (vdp->is_hires) {
	    a2v_render_hires(vdp);
	} else {
	    a2v_render_lores(vdp);
	}
    } else if (vdp->scanline == 262) {
	video_display_buffer();
	vdp->scanline = 0;
	return;
    }
    
    vdp->scanline++;
}

void a2v_load_chargen(apple2_vdp vdp)
{
    rom_file romfile;

    vdp->chargen = NULL;
    
    romfile = read_romimage("applecg.bin");
    
    if (!romfile) {
	deb_printf("unable to load character generator.\n");
	return;
    }

    vdp->chargen = romfile->data;

    /* FIXME: memory leak */
}

apple2_vdp a2v_init(u8 *system_memory)
{
    apple2_vdp retval;

    retval = malloc(sizeof(struct apple2_vdp));

    if (!retval) {
	return NULL;
    }

    a2v_load_chargen(retval);

    if (!retval->chargen) {
	free(retval);
	return NULL;
    }
    
    retval->memory = system_memory;

    video_setsize(280, 192);

    video_setpal(8, a2v_palbase_red, a2v_palbase_green, a2v_palbase_blue);

    return retval;
}

/*
 * $Log: apple2_vdp.c,v $
 * Revision 1.1  2000/01/01 04:12:20  nyef
 * Initial revision
 *
 */
