/** M6280: portable 6280 emulator ****************************/
/**                                                         **/
/**                         M6280.c                         **/
/**                                                         **/
/** This file contains implementation for 6280 CPU. Don't   **/
/** forget to provide Rd6280(), Wr6280(), Loop6280(), and   **/
/** possibly Op6280() functions to accomodate the emulated  **/
/** machine's architecture.                                 **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996                      **/
/**               Alex Krasivsky  1996                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/   
/**     changes to this file.                               **/
/*************************************************************/

/*
 * this file has been modified for use with the cal.
 */

/* $Id: m6280.c,v 1.2 1999/08/04 00:47:17 nyef Exp $ */

#include "m6280.h"
#include "ui.h"
#include "tool.h"
#include <stdio.h>

static byte Cycles[256] = /* FIXME: very wrong */
{
  7,6,2,8,3,3,5,5,3,2,2,2,4,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7,
  6,6,2,8,3,3,5,5,4,2,2,2,4,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7,
  6,6,2,8,3,3,5,5,3,2,2,2,3,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7,
  6,6,2,8,3,3,5,5,4,2,2,2,5,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7,
  2,6,2,6,3,3,3,3,2,2,2,2,4,4,4,4,
  2,6,2,6,4,4,4,4,2,5,2,5,5,5,5,5,
  2,6,2,6,3,3,3,3,2,2,2,2,4,4,4,4,
  2,5,2,5,4,4,4,4,2,4,2,5,4,4,4,4,
  2,6,2,8,3,3,5,5,2,2,2,2,4,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7,
  2,6,2,8,3,3,5,5,2,2,2,2,4,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7
};

extern byte ZNTable[256];

/** INLINE ***************************************************/
/** Different compilers inline C functions differently.     **/
/*************************************************************/
#ifdef __GNUC__
#define INLINE inline
#else
#define INLINE static
#endif

/** System-Dependent Stuff ***********************************/
/** This is system-dependent code put here to speed things  **/
/** up. It has to stay inlined to be fast.                  **/
/*************************************************************/
INLINE byte Rd6280(M6280 *R, word addr) { return (R->readtable[(addr >> R->memshift) & R->memmask](R->User, addr)); }
INLINE void Wr6280(M6280 *R, word addr, byte data) { (R->writetable[(addr >> R->memshift) & R->memmask](R->User, addr, data)); }

/** FAST_RDOP ************************************************/
/** With this #define not present, Rd6280() should perform  **/
/** the functions of Rd6280().                              **/
/*************************************************************/
#ifndef FAST_RDOP
#define Op6280(R, A) Rd6280(R, A)
#endif

/** Addressing Methods ***************************************/
/** These macros calculate and return effective addresses.  **/
/*************************************************************/
#define MC_Ab(Rg)	M_LDWORD(Rg)
#define MC_Zp(Rg)	Rg.B.l=Op6280(R, R->PC.W++);Rg.B.h=0
#define MC_Zx(Rg)	Rg.B.l=Op6280(R, R->PC.W++)+R->X;Rg.B.h=0
#define MC_Zy(Rg)	Rg.B.l=Op6280(R, R->PC.W++)+R->Y;Rg.B.h=0
#define MC_Ax(Rg)	M_LDWORD(Rg);Rg.W+=R->X
#define MC_Ay(Rg)	M_LDWORD(Rg);Rg.W+=R->Y
#define MC_Id(Rg)	K.B.l=Op6280(R, R->PC.W++);K.B.h=0; \
			Rg.B.l=R->page_0[K.W++];Rg.B.h=R->page_0[K.W]
#define MC_Ix(Rg)	K.B.l=Op6280(R, R->PC.W++)+R->X;K.B.h=0; \
			Rg.B.l=R->page_0[K.W++];Rg.B.h=R->page_0[K.W]
#define MC_Iy(Rg)	K.B.l=Op6280(R, R->PC.W++);K.B.h=0; \
			Rg.B.l=R->page_0[K.W++];Rg.B.h=R->page_0[K.W]; \
			Rg.W+=R->Y

/** Reading From Memory **************************************/
/** These macros calculate address and read from it.        **/
/*************************************************************/
#define MR_Ab(Rg)	MC_Ab(J);Rg=Rd6280(R, J.W)
#define MR_Im(Rg)	Rg=Op6280(R, R->PC.W++)
#define	MR_Zp(Rg)	MC_Zp(J);Rg=R->page_0[J.W];
#define MR_Zx(Rg)	MC_Zx(J);Rg=R->page_0[J.W];
#define MR_Zy(Rg)	MC_Zy(J);Rg=R->page_0[J.W];
#define	MR_Ax(Rg)	MC_Ax(J);Rg=Rd6280(R, J.W)
#define MR_Ay(Rg)	MC_Ay(J);Rg=Rd6280(R, J.W)
#define MR_Id(Rg)	MC_Id(J);Rg=Rd6280(R, J.W)
#define MR_Ix(Rg)	MC_Ix(J);Rg=Rd6280(R, J.W)
#define MR_Iy(Rg)	MC_Iy(J);Rg=Rd6280(R, J.W)

/** Writing To Memory ****************************************/
/** These macros calculate address and write to it.         **/
/*************************************************************/
#define MW_Ab(Rg)	MC_Ab(J);Wr6280(R, J.W,Rg)
#define MW_Zp(Rg)	MC_Zp(J);R->page_0[J.W]=Rg;
#define MW_Zx(Rg)	MC_Zx(J);R->page_0[J.W]=Rg;
#define MW_Zy(Rg)	MC_Zy(J);R->page_0[J.W]=Rg;
#define MW_Ax(Rg)	MC_Ax(J);Wr6280(R, J.W,Rg)
#define MW_Ay(Rg)	MC_Ay(J);Wr6280(R, J.W,Rg)
#define MW_Ix(Rg)	MC_Ix(J);Wr6280(R, J.W,Rg)
#define MW_Iy(Rg)	MC_Iy(J);Wr6280(R, J.W,Rg)

/** Modifying Memory *****************************************/
/** These macros calculate address and modify it.           **/
/*************************************************************/
#define MM_Ab(Cmd)	MC_Ab(J);I=Rd6280(R, J.W);Cmd(I);Wr6280(R, J.W,I)
#define MM_Zp(Cmd)	MC_Zp(J);I=R->page_0[J.W];Cmd(I);R->page_0[J.W]=I
#define MM_Zx(Cmd)	MC_Zx(J);I=R->page_0[J.W];Cmd(I);R->page_0[J.W]=I
#define MM_Ax(Cmd)	MC_Ax(J);I=Rd6280(R, J.W);Cmd(I);Wr6280(R, J.W,I)

/** Other Macros *********************************************/
/** Calculating flags, stack, jumps, arithmetics, etc.      **/
/*************************************************************/
#define M_FL(Rg)	R->P=(R->P&~(M6280_Z_FLAG|M6280_N_FLAG))|ZNTable[Rg]
#define M_LDWORD(Rg)	Rg.B.l=Op6280(R, R->PC.W++);Rg.B.h=Op6280(R, R->PC.W++)

#define M_PUSH(Rg)	R->page_0[0x0100|R->S] = Rg;R->S--
#define M_POP(Rg)	R->S++;Rg=R->page_0[0x0100|R->S]
#define M_JR		R->PC.W+=(offset)Op6280(R, R->PC.W)+1;R->ICount--

#define M_ADC(Rg) \
  if(R->P&M6280_D_FLAG) \
  { \
    K.B.l=(R->A&0x0F)+(Rg&0x0F)+(R->P&M6280_C_FLAG); \
    K.B.h=(R->A>>4)+(Rg>>4)/* +(K.B.l>15? 1:0) */; \
    if(K.B.l>9) { K.B.l+=6;K.B.h++; } \
    if(K.B.h>9) K.B.h+=6; \
    R->A=(K.B.l&0x0F)|(K.B.h<<4); \
    R->P=(R->P&~M6280_C_FLAG)|(K.B.h>15? M6280_C_FLAG:0); \
    M_FL(R->A); \
  } \
  else \
  { \
    K.W=R->A+Rg+(R->P&M6280_C_FLAG); \
    R->P&=~(M6280_N_FLAG|M6280_V_FLAG|M6280_Z_FLAG|M6280_C_FLAG); \
    R->P|=(~(R->A^Rg)&(R->A^K.B.l)&0x80? M6280_V_FLAG:0)| \
          (K.B.h? M6280_C_FLAG:0)|ZNTable[K.B.l]; \
    R->A=K.B.l; \
  }

/* Warning! M6280_C_FLAG is inverted before SBC and after it */
#define M_SBC(Rg) \
  if(R->P&M6280_D_FLAG) \
  { \
    K.B.l=(R->A&0x0F)-(Rg&0x0F)-(~R->P&M6280_C_FLAG); \
    if(K.B.l&0x10) K.B.l-=6; \
    K.B.h=(R->A>>4)-(Rg>>4)-((K.B.l&0x10)==0x10); \
    if(K.B.h&0x10) K.B.h-=6; \
    R->A=(K.B.l&0x0F)|(K.B.h<<4); \
    R->P=(R->P&~M6280_C_FLAG)|(K.B.h>15? 0:M6280_C_FLAG); \
    M_FL(R->A); \
  } \
  else \
  { \
    K.W=R->A-Rg-(~R->P&M6280_C_FLAG); \
    R->P&=~(M6280_N_FLAG|M6280_V_FLAG|M6280_Z_FLAG|M6280_C_FLAG); \
    R->P|=((R->A^Rg)&(R->A^K.B.l)&0x80? M6280_V_FLAG:0)| \
          (K.B.h? 0:M6280_C_FLAG)|ZNTable[K.B.l]; \
    R->A=K.B.l; \
  }

#define M_CMP(Rg1,Rg2) \
  K.W=Rg1-Rg2; \
  R->P&=~(M6280_N_FLAG|M6280_Z_FLAG|M6280_C_FLAG); \
  R->P|=ZNTable[K.B.l]|(K.B.h? 0:M6280_C_FLAG)
#define M_BIT(Rg) \
  R->P&=~(M6280_N_FLAG|M6280_V_FLAG|M6280_Z_FLAG); \
  R->P|=(Rg&(M6280_N_FLAG|M6280_V_FLAG))|(Rg&R->A? 0:M6280_Z_FLAG)

#define M_AND(Rg)	R->A&=Rg;M_FL(R->A)
#define M_ORA(Rg)	R->A|=Rg;M_FL(R->A)
#define M_EOR(Rg)	R->A^=Rg;M_FL(R->A)
#define M_INC(Rg)	Rg++;M_FL(Rg)
#define M_DEC(Rg)	Rg--;M_FL(Rg)

#define M_ASL(Rg)	R->P&=~M6280_C_FLAG;R->P|=Rg>>7;Rg<<=1;M_FL(Rg)
#define M_LSR(Rg)	R->P&=~M6280_C_FLAG;R->P|=Rg&M6280_C_FLAG;Rg>>=1;M_FL(Rg)
#define M_ROL(Rg)	K.B.l=(Rg<<1)|(R->P&M6280_C_FLAG); \
			R->P&=~M6280_C_FLAG;R->P|=Rg>>7;Rg=K.B.l; \
			M_FL(Rg)
#define M_ROR(Rg)	K.B.l=(Rg>>1)|(R->P<<7); \
			R->P&=~M6280_C_FLAG;R->P|=Rg&M6280_C_FLAG;Rg=K.B.l; \
			M_FL(Rg)

/** Reset6280() **********************************************/
/** This function can be used to reset the registers before **/
/** starting execution with Run6280(). It sets registers to **/
/** their initial values.                                   **/
/*************************************************************/
void Reset6280(M6280 *R)
{
  R->A=R->X=R->Y=0x00;
  R->P=M6280_Z_FLAG|M6280_R_FLAG;
  R->S=0xFF;
  R->PC.B.l=Rd6280(R, 0xFFFE);
  R->PC.B.h=Rd6280(R, 0xFFFF);
  R->ICount=R->IPeriod;
  R->IRequest=INT_NONE;
  R->AfterCLI=0;
}

/** Int6280() ************************************************/
/** This function will generate interrupt of a given type.  **/
/** INT_NMI will cause a non-maskable interrupt. INT_IRQ    **/
/** will cause a normal interrupt, unless M6280_I_FLAG set in R.  **/
/*************************************************************/
void Int6280(M6280 *R,word Type)
{
  register pair J;

  if((Type==INT_NMI)||((Type==INT_IRQ)&&!(R->P&M6280_I_FLAG)))
  {
    R->ICount-=7;
    M_PUSH(R->PC.B.h);
    M_PUSH(R->PC.B.l);
    M_PUSH(R->P&~M6280_B_FLAG);
    R->P&=~M6280_D_FLAG;
    if(Type==INT_NMI) J.W=0xFFFE; else { R->P|=M6280_I_FLAG;J.W=0xFFF8; }
    R->PC.B.l=Rd6280(R, J.W++);
    R->PC.B.h=Rd6280(R, J.W);
  }
}

/** Run6280() ************************************************/
/** This function will run 6280 code until LoOp6280(R, ) call  **/
/** returns INT_QUIT. It will return the PC at which        **/
/** emulation stopped, and current register values in R.    **/
/*************************************************************/
word Run6280(M6280 *R)
{
    register pair J;
    register pair K;
    register byte I;

    J.W = 0; /* gcc -O2 is a chode */
    K.W = 0; /* gcc -O2 is a chode */
    
    R->ICount+=R->IPeriod;     /* Reset the cycle counter   */
    
    while ((R->ICount > 0) || (R->AfterCLI)) {
	K.W = R->IRequest;
	R->IRequest = INT_NONE;
	
	/* If we have come after CLI, get INT_? from IRequest */
	/* Otherwise, get it from the loop handler            */
	if(R->AfterCLI) {
	    R->ICount+=R->IBackup-1;  /* Restore the ICount        */
	    R->AfterCLI=0;            /* Done with AfterCLI state  */
	}
	
	if(K.W==INT_QUIT) return(R->PC.W); /* Exit if INT_QUIT     */
	if(K.W!=INT_NONE) Int6280(R,K.W);  /* Interrupt if needed  */

/* 	deb_printf("PC: 0x%04hx.\n", R->PC); */
	I=Op6280(R, R->PC.W++);
	R->ICount-=Cycles[I];
	switch (I) {
#include "m6280opc.h"
	}
    }
    
    /* Execution stopped */
    return(R->PC.W);
}

/*
 * $Log: m6280.c,v $
 * Revision 1.2  1999/08/04 00:47:17  nyef
 * fixed some problems with the ADC and SBC instructions
 *
 * Revision 1.1  1999/08/02 00:19:19  nyef
 * Initial revision
 *
 */
