/** M6502: portable 6502 emulator ****************************/
/**                                                         **/
/**                         M6502.c                         **/
/**                                                         **/
/** This file contains implementation for 6502 CPU. Don't   **/
/** forget to provide Rd6502(), Wr6502(), Loop6502(), and   **/
/** possibly Op6502() functions to accomodate the emulated  **/
/** machine's architecture.                                 **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996                      **/
/**               Alex Krasivsky  1996                      **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/   
/**     changes to this file.                               **/
/*************************************************************/

/*
 * this file has been modified for use with the cal.
 */

/* $Id: m6502.c,v 1.5 1999/08/04 01:10:29 nyef Exp $ */

#include "m6502.h"
#include "ui.h"
#include <stdio.h>

static byte Cycles[256] =
{
  7,6,2,8,3,3,5,5,3,2,2,2,4,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7,
  6,6,2,8,3,3,5,5,4,2,2,2,4,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7,
  6,6,2,8,3,3,5,5,3,2,2,2,3,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7,
  6,6,2,8,3,3,5,5,4,2,2,2,5,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7,
  2,6,2,6,3,3,3,3,2,2,2,2,4,4,4,4,
  2,6,2,6,4,4,4,4,2,5,2,5,5,5,5,5,
  2,6,2,6,3,3,3,3,2,2,2,2,4,4,4,4,
  2,5,2,5,4,4,4,4,2,4,2,5,4,4,4,4,
  2,6,2,8,3,3,5,5,2,2,2,2,4,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7,
  2,6,2,8,3,3,5,5,2,2,2,2,4,4,6,6,
  2,5,2,8,4,4,6,6,2,4,2,7,5,5,7,7
};

byte ZNTable[256] =
{
  M6502_Z_FLAG,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
  M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,M6502_N_FLAG,
};

/** INLINE ***************************************************/
/** Different compilers inline C functions differently.     **/
/*************************************************************/
#ifdef __GNUC__
#define INLINE inline
#else
#define INLINE static
#endif

/** System-Dependent Stuff ***********************************/
/** This is system-dependent code put here to speed things  **/
/** up. It has to stay inlined to be fast.                  **/
/*************************************************************/
INLINE byte Rd6502(M6502 *R, word addr) { return (R->readtable[(addr >> R->memshift) & R->memmask](R->User, addr)); }
INLINE void Wr6502(M6502 *R, word addr, byte data) { (R->writetable[(addr >> R->memshift) & R->memmask](R->User, addr, data)); }

/** FAST_RDOP ************************************************/
/** With this #define not present, Rd6502() should perform  **/
/** the functions of Rd6502().                              **/
/*************************************************************/
#ifndef FAST_RDOP
#define Op6502(R, A) Rd6502(R, A)
#endif

/** Addressing Methods ***************************************/
/** These macros calculate and return effective addresses.  **/
/*************************************************************/
#define MC_Ab(Rg)	M_LDWORD(Rg)
#define MC_Zp(Rg)	Rg.B.l=Op6502(R, R->PC.W++);Rg.B.h=0
#define MC_Zx(Rg)	Rg.B.l=Op6502(R, R->PC.W++)+R->X;Rg.B.h=0
#define MC_Zy(Rg)	Rg.B.l=Op6502(R, R->PC.W++)+R->Y;Rg.B.h=0
#define MC_Ax(Rg)	M_LDWORD(Rg);Rg.W+=R->X
#define MC_Ay(Rg)	M_LDWORD(Rg);Rg.W+=R->Y
#define MC_Ix(Rg)	K.B.l=Op6502(R, R->PC.W++)+R->X;K.B.h=0; \
			Rg.B.l=Op6502(R, K.W++);Rg.B.h=Op6502(R, K.W)
#define MC_Iy(Rg)	K.B.l=Op6502(R, R->PC.W++);K.B.h=0; \
			Rg.B.l=Op6502(R, K.W++);Rg.B.h=Op6502(R, K.W); \
			Rg.W+=R->Y

/** Reading From Memory **************************************/
/** These macros calculate address and read from it.        **/
/*************************************************************/
#define MR_Ab(Rg)	MC_Ab(J);Rg=Rd6502(R, J.W)
#define MR_Im(Rg)	Rg=Op6502(R, R->PC.W++)
#define	MR_Zp(Rg)	MC_Zp(J);Rg=Rd6502(R, J.W)
#define MR_Zx(Rg)	MC_Zx(J);Rg=Rd6502(R, J.W)
#define MR_Zy(Rg)	MC_Zy(J);Rg=Rd6502(R, J.W)
#define	MR_Ax(Rg)	MC_Ax(J);Rg=Rd6502(R, J.W)
#define MR_Ay(Rg)	MC_Ay(J);Rg=Rd6502(R, J.W)
#define MR_Ix(Rg)	MC_Ix(J);Rg=Rd6502(R, J.W)
#define MR_Iy(Rg)	MC_Iy(J);Rg=Rd6502(R, J.W)

/** Writing To Memory ****************************************/
/** These macros calculate address and write to it.         **/
/*************************************************************/
#define MW_Ab(Rg)	MC_Ab(J);Wr6502(R, J.W,Rg)
#define MW_Zp(Rg)	MC_Zp(J);Wr6502(R, J.W,Rg)
#define MW_Zx(Rg)	MC_Zx(J);Wr6502(R, J.W,Rg)
#define MW_Zy(Rg)	MC_Zy(J);Wr6502(R, J.W,Rg)
#define MW_Ax(Rg)	MC_Ax(J);Wr6502(R, J.W,Rg)
#define MW_Ay(Rg)	MC_Ay(J);Wr6502(R, J.W,Rg)
#define MW_Ix(Rg)	MC_Ix(J);Wr6502(R, J.W,Rg)
#define MW_Iy(Rg)	MC_Iy(J);Wr6502(R, J.W,Rg)

/** Modifying Memory *****************************************/
/** These macros calculate address and modify it.           **/
/*************************************************************/
#define MM_Ab(Cmd)	MC_Ab(J);I=Rd6502(R, J.W);Cmd(I);Wr6502(R, J.W,I)
#define MM_Zp(Cmd)	MC_Zp(J);I=Rd6502(R, J.W);Cmd(I);Wr6502(R, J.W,I)
#define MM_Zx(Cmd)	MC_Zx(J);I=Rd6502(R, J.W);Cmd(I);Wr6502(R, J.W,I)
#define MM_Ax(Cmd)	MC_Ax(J);I=Rd6502(R, J.W);Cmd(I);Wr6502(R, J.W,I)

/** Other Macros *********************************************/
/** Calculating flags, stack, jumps, arithmetics, etc.      **/
/*************************************************************/
#define M_FL(Rg)	R->P=(R->P&~(M6502_Z_FLAG|M6502_N_FLAG))|ZNTable[Rg]
#define M_LDWORD(Rg)	Rg.B.l=Op6502(R, R->PC.W++);Rg.B.h=Op6502(R, R->PC.W++)

#define M_PUSH(Rg)	Wr6502(R, 0x0100|R->S,Rg);R->S--
#define M_POP(Rg)	R->S++;Rg=Op6502(R, 0x0100|R->S)
#define M_JR		R->PC.W+=(offset)Op6502(R, R->PC.W)+1;R->ICount--

#define M_ADC(Rg) \
  if(R->P&M6502_D_FLAG) \
  { \
    K.B.l=(R->A&0x0F)+(Rg&0x0F)+(R->P&M6502_C_FLAG); \
    K.B.h=(R->A>>4)+(Rg>>4)/* +(K.B.l>15? 1:0) */; \
    if(K.B.l>9) { K.B.l+=6;K.B.h++; } \
    if(K.B.h>9) K.B.h+=6; \
    R->A=(K.B.l&0x0F)|(K.B.h<<4); \
    R->P=(R->P&~M6502_C_FLAG)|(K.B.h>15? M6502_C_FLAG:0); \
    M_FL(R->A); \
  } \
  else \
  { \
    K.W=R->A+Rg+(R->P&M6502_C_FLAG); \
    R->P&=~(M6502_N_FLAG|M6502_V_FLAG|M6502_Z_FLAG|M6502_C_FLAG); \
    R->P|=(~(R->A^Rg)&(R->A^K.B.l)&0x80? M6502_V_FLAG:0)| \
          (K.B.h? M6502_C_FLAG:0)|ZNTable[K.B.l]; \
    R->A=K.B.l; \
  }

/* Warning! M6502_C_FLAG is inverted before SBC and after it */
#define M_SBC(Rg) \
  if(R->P&M6502_D_FLAG) \
  { \
    K.B.l=(R->A&0x0F)-(Rg&0x0F)-(~R->P&M6502_C_FLAG); \
    if(K.B.l&0x10) K.B.l-=6; \
    K.B.h=(R->A>>4)-(Rg>>4)-((K.B.l&0x10)==0x10); \
    if(K.B.h&0x10) K.B.h-=6; \
    R->A=(K.B.l&0x0F)|(K.B.h<<4); \
    R->P=(R->P&~M6502_C_FLAG)|(K.B.h>15? 0:M6502_C_FLAG); \
    M_FL(R->A); \
  } \
  else \
  { \
    K.W=R->A-Rg-(~R->P&M6502_C_FLAG); \
    R->P&=~(M6502_N_FLAG|M6502_V_FLAG|M6502_Z_FLAG|M6502_C_FLAG); \
    R->P|=((R->A^Rg)&(R->A^K.B.l)&0x80? M6502_V_FLAG:0)| \
          (K.B.h? 0:M6502_C_FLAG)|ZNTable[K.B.l]; \
    R->A=K.B.l; \
  }

#define M_CMP(Rg1,Rg2) \
  K.W=Rg1-Rg2; \
  R->P&=~(M6502_N_FLAG|M6502_Z_FLAG|M6502_C_FLAG); \
  R->P|=ZNTable[K.B.l]|(K.B.h? 0:M6502_C_FLAG)
#define M_BIT(Rg) \
  R->P&=~(M6502_N_FLAG|M6502_V_FLAG|M6502_Z_FLAG); \
  R->P|=(Rg&(M6502_N_FLAG|M6502_V_FLAG))|(Rg&R->A? 0:M6502_Z_FLAG)

#define M_AND(Rg)	R->A&=Rg;M_FL(R->A)
#define M_ORA(Rg)	R->A|=Rg;M_FL(R->A)
#define M_EOR(Rg)	R->A^=Rg;M_FL(R->A)
#define M_INC(Rg)	Rg++;M_FL(Rg)
#define M_DEC(Rg)	Rg--;M_FL(Rg)

#define M_ASL(Rg)	R->P&=~M6502_C_FLAG;R->P|=Rg>>7;Rg<<=1;M_FL(Rg)
#define M_LSR(Rg)	R->P&=~M6502_C_FLAG;R->P|=Rg&M6502_C_FLAG;Rg>>=1;M_FL(Rg)
#define M_ROL(Rg)	K.B.l=(Rg<<1)|(R->P&M6502_C_FLAG); \
			R->P&=~M6502_C_FLAG;R->P|=Rg>>7;Rg=K.B.l; \
			M_FL(Rg)
#define M_ROR(Rg)	K.B.l=(Rg>>1)|(R->P<<7); \
			R->P&=~M6502_C_FLAG;R->P|=Rg&M6502_C_FLAG;Rg=K.B.l; \
			M_FL(Rg)

/** Reset6502() **********************************************/
/** This function can be used to reset the registers before **/
/** starting execution with Run6502(). It sets registers to **/
/** their initial values.                                   **/
/*************************************************************/
void Reset6502(M6502 *R)
{
  R->A=R->X=R->Y=0x00;
  R->P=M6502_Z_FLAG|M6502_R_FLAG;
  R->S=0xFF;
  R->PC.B.l=Rd6502(R, 0xFFFC);
  R->PC.B.h=Rd6502(R, 0xFFFD);   
  R->ICount=R->IPeriod;
  R->IRequest=INT_NONE;
  R->AfterCLI=0;
}

/** Int6502() ************************************************/
/** This function will generate interrupt of a given type.  **/
/** INT_NMI will cause a non-maskable interrupt. INT_IRQ    **/
/** will cause a normal interrupt, unless M6502_I_FLAG set in R.  **/
/*************************************************************/
void Int6502(M6502 *R,word Type)
{
  register pair J;

  if((Type==INT_NMI)||((Type==INT_IRQ)&&!(R->P&M6502_I_FLAG)))
  {
    R->ICount-=7;
    M_PUSH(R->PC.B.h);
    M_PUSH(R->PC.B.l);
    M_PUSH(R->P&~M6502_B_FLAG);
    R->P&=~M6502_D_FLAG;
    if(Type==INT_NMI) J.W=0xFFFA; else { R->P|=M6502_I_FLAG;J.W=0xFFFE; }
    R->PC.B.l=Rd6502(R, J.W++);
    R->PC.B.h=Rd6502(R, J.W);
  }
}

/** Run6502() ************************************************/
/** This function will run 6502 code until LoOp6502(R, ) call  **/
/** returns INT_QUIT. It will return the PC at which        **/
/** emulation stopped, and current register values in R.    **/
/*************************************************************/
word Run6502(M6502 *R)
{
    register pair J;
    register pair K;
    register byte I;

    J.W = 0; /* gcc -O2 is a chode */
    K.W = 0; /* gcc -O2 is a chode */
    
    R->ICount+=R->IPeriod;     /* Reset the cycle counter   */
    
    while ((R->ICount > 0) || (R->AfterCLI)) {
	K.W = R->IRequest;
	R->IRequest = INT_NONE;
	
	/* If we have come after CLI, get INT_? from IRequest */
	/* Otherwise, get it from the loop handler            */
	if(R->AfterCLI) {
	    R->ICount+=R->IBackup-1;  /* Restore the ICount        */
	    R->AfterCLI=0;            /* Done with AfterCLI state  */
	}
	
	if(K.W==INT_QUIT) return(R->PC.W); /* Exit if INT_QUIT     */
	if(K.W!=INT_NONE) Int6502(R,K.W);  /* Interrupt if needed  */
	
	I=Op6502(R, R->PC.W++);
	R->ICount-=Cycles[I];
	switch (I) {
#include "m6502opc.h"
	}
    }
    
    /* Execution stopped */
    return(R->PC.W);
}

/*
 * $Log: m6502.c,v $
 * Revision 1.5  1999/08/04 01:10:29  nyef
 * fixed some problems with the ADC and SBC instructions
 *
 * Revision 1.4  1999/07/21 00:35:02  nyef
 * included ui.h to squash warning with debug code in m6502opc.h
 *
 * Revision 1.3  1999/01/03 02:24:01  nyef
 * modified interrupt handling to not conflict with mz80.
 *
 * Revision 1.2  1998/12/22 04:57:08  nyef
 * changed x_FLAG to M6502_x_FLAG.
 *
 * Revision 1.1  1998/12/16 03:50:23  nyef
 * Initial revision
 *
 */
