/*
 * ui.h
 *
 * user interface interface
 */

/* $Id: ui.h,v 1.5 2000/01/01 03:22:52 nyef Exp $ */

#ifndef __UI_H__
#define __UI_H__

#include "types.h"

/* debug console interface */

void deb_printf(const char *fmt, ...);


/* timeslicer interface */

void set_timeslice(void (*proc)(void *), void *data);
void unset_timeslice(void);


/* shutdown callback */

typedef void (*shutdown_t)(void);
extern shutdown_t dn_shutdown;


/* joypad interface */

/*
 * joypad buttons are: UP, DOWN, LEFT, RIGHT, whatever,
 *    START (opt), SELECT (opt)
 */

struct joypad_template {
    int num_buttons;
    unsigned long buttons[0];
};

struct joypad {
    const struct joypad_template *template;
    unsigned long data;
};

int ui_register_joypad(struct joypad *pad);
void ui_update_joypad(struct joypad *pad);

/* keypad interface */

struct keypad {
    u16 data;
};

int keypad_register(struct keypad *pad);
void keypad_update(struct keypad *pad);

#define KPD_0 0x0001
#define KPD_1 0x0002
#define KPD_2 0x0004
#define KPD_3 0x0008
#define KPD_4 0x0010
#define KPD_5 0x0020
#define KPD_6 0x0040
#define KPD_7 0x0080
#define KPD_8 0x0100
#define KPD_9 0x0200
#define KPD_STAR 0x0400
#define KPD_HASH 0x0800
#define KPD_A 0x1000 /* Unused */
#define KPD_B 0x2000 /* Unused */
#define KPD_C 0x4000 /* Unused */
#define KPD_D 0x8000 /* Unused */

#endif /* __UI_H__ */

/*
 * $Log: ui.h,v $
 * Revision 1.5  2000/01/01 03:22:52  nyef
 * added preliminary keypad interface
 *
 * Revision 1.4  1999/06/05 02:42:04  nyef
 * added new joypad interface
 *
 * Revision 1.3  1999/04/17 20:11:27  nyef
 * changed shutdown() to dn_shutdown().
 *
 * Revision 1.2  1998/12/21 02:58:13  nyef
 * added a shutdown callback.
 *
 * Revision 1.1  1998/08/02 04:13:25  nyef
 * Initial revision
 *
 */
