/*
 * video_x.c
 *
 * display screen management
 */

/* $Id: video_x.c,v 1.18 2000/01/01 03:22:38 nyef Exp $ */

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/extensions/XShm.h>
#include <sys/shm.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include "video.h"
#include "tool.h"
#include "ui.h"

/* the joypad (only one, I could care less about more) */
struct joypad *ui_joypad;

/* the keypad (only one, I could care less about more) */
struct keypad *ui_keypad;

extern int errno;

int video_active = 0;
int deblevel = 0;

Display *display;
Window nes_window;
GC nes_gc;
XImage *nes_image_0;
XImage *nes_image_1;
XImage *nes_image;
Colormap nes_cmap;
Screen *nes_screen;

int buffer_x;
int buffer_y;
char *image_buf_0;
char *image_buf_1;
char *image_buf;

typedef void ((*xlatfunc)(void *, int));

vbpfunc video_get_vbp;
unsigned char *vbpfunc8(int line);
unsigned char *vbpfunc16(int line);
xlatfunc vidxlate;
void vidxlate16 (char *dest, int size);
video_setpal_t video_setpal;
void init_xlate8(int colors, int *red, int *green, int *blue);
void init_xlate16(int colors, int *red, int *green, int *blue);
unsigned char *vid_pre_xlat;

/* FIXME: ui_x.c interface -- move to separate header. */
extern Widget drawbox;
extern Widget manager;
void ui_set_drawbox_size(int width, int height);

void do_key_event(Widget, XtPointer, XEvent *, Boolean *);

/*
 * Thanks to Sam Lantinga of SDL and the GDK team for XShm references.
 */
int use_shm = False;
int shm_error = False;
int (*old_x_handler)(Display *, XErrorEvent *);
int alloc_shm_surface(char**, XImage**, int, int, int);
int shm_error_handler(Display*, XErrorEvent *);

void video_init()
{
    int is_local_display;

    display = XtDisplay(drawbox);
    nes_screen = XtScreen(drawbox);
    nes_window = XtWindow(drawbox);

    ui_set_drawbox_size(256, 240);

    nes_gc = DefaultGCOfScreen(nes_screen);
    XtVaGetValues(drawbox, XtNcolormap, &nes_cmap, NULL);

    XtAddEventHandler(manager, KeyPressMask | KeyReleaseMask, False, do_key_event, NULL);
    
    XFlush(display);

    if (0 == strncmp(XDisplayName(NULL), ":", 1)) {
	is_local_display = True;
    } else {
	is_local_display = False;
    }

    if (is_local_display) {
	use_shm = XShmQueryExtension(display);
    } else {
	use_shm = False;
    }

#ifdef FORCE_NOXSHM
    use_shm = False;
#endif
    
    if (DefaultDepthOfScreen(nes_screen) == 8) {
	video_setpal = init_xlate8;
    } else if (DefaultDepthOfScreen(nes_screen) == 16) {
	video_setpal = init_xlate16;
    } else {
	fprintf(stderr, "display depth %d unsupported.\n", DefaultDepthOfScreen(nes_screen));
	exit(1);
    }
}

void video_shutdown(void)
{
}

void video_setsize(int x, int y)
{
    buffer_x = x;
    buffer_y = y;

    ui_set_drawbox_size(x, y);
    
    if (DefaultDepthOfScreen(nes_screen) == 8) {

	if (use_shm) {
	    if (False == alloc_shm_surface(&image_buf_0, &nes_image_0, 8, x, y)) {
		fprintf(stderr, "darcnes: could not alloc_shm_surface(8)\n");
		exit( 1 );
	    }
	    if (False == alloc_shm_surface(&image_buf_1, &nes_image_1, 8, x, y)) {
		fprintf(stderr, "darcnes: could not alloc_shm_surface(8)\n");
		exit( 1 );
	    }
	} else {
	    image_buf_0 = malloc(x * y);
	    nes_image_0 = XCreateImage(display, DefaultVisualOfScreen(nes_screen),
				     8, ZPixmap, 0, image_buf_0, x, y, 8, 0);
	    image_buf_1 = malloc(x * y);
	    nes_image_1 = XCreateImage(display, DefaultVisualOfScreen(nes_screen),
				     8, ZPixmap, 0, image_buf_1, x, y, 8, 0);
	}
    } else if (DefaultDepthOfScreen(nes_screen) == 16) {
	
	if (use_shm) {
	    if (False == alloc_shm_surface(&image_buf_0, &nes_image_0, 16, x, y)) {
		fprintf (stderr, "darcnes: could not alloc_shm_surface(16)\n");
		exit(1);
	    }
	    if (False == alloc_shm_surface(&image_buf_1, &nes_image_1, 16, x, y)) {
		fprintf (stderr, "darcnes: could not alloc_shm_surface(16)\n");
		exit(1);
	    }
	} else {
	    image_buf_0 = malloc(x * y * 2);
	    nes_image_0 = XCreateImage(display, DefaultVisualOfScreen(nes_screen),
				     16, ZPixmap, 0, image_buf_0, x, y, 8, 0);
	    image_buf_1 = malloc(x * y * 2);
	    nes_image_1 = XCreateImage(display, DefaultVisualOfScreen(nes_screen),
				     16, ZPixmap, 0, image_buf_1, x, y, 8, 0);
	}
    }

    image_buf = image_buf_0;
    nes_image = nes_image_0;
}

void video_enter_deb(void)
{
    if (video_active && (!deblevel)) {
    }
    deblevel++;
}

int shm_error_handler(Display* d, XErrorEvent* e)
{
    if (BadAccess == e->error_code) {
	shm_error = True;
	return 0;
    }

    return old_x_handler(d, e);
}

int alloc_shm_surface(char** buffer, XImage** image, int depth, int width, int height)
{
    XShmSegmentInfo* shm_info;
    int byte_depth = depth / 8;
    Screen *screen = XtScreen(drawbox);

    shm_info = (XShmSegmentInfo*) malloc(sizeof(XShmSegmentInfo));

    if (!use_shm) {
	fprintf(stderr, "darcnes: alloc_shm_surface() when use_shm == False\n");
	return False;
    }

    shm_info->shmid = shmget(IPC_PRIVATE, width * height * byte_depth,
			     IPC_CREAT | 0777);

    if (shm_info->shmid < 0) {
	fprintf(stderr, "darcnes: shmget() failed\n");
	use_shm = False;
	return False;
    }

    shm_info->shmaddr = (char *)shmat(shm_info->shmid, 0, 0);
    shm_info->readOnly = False;
 
    if ((char *) -1 == shm_info->shmaddr) {
	fprintf(stderr, "darcnes: shmat() failed: ");

	switch (errno) {
	case EACCES:
	    fprintf(stderr, "EACCES\n");
	    break;
	case EINVAL:
	    fprintf(stderr, "EINVAL\n");
	    break;
	case ENOMEM:
	    fprintf(stderr, "ENOMEM\n");
	    break;
	default:
	    fprintf(stderr, "Unknown\n");
	    break;
	}

	use_shm = False;
	return False;
    }

    shm_error = False;
    /*
     * We set this to grab anything bad that happens during the
     * attach and (especially) sync calls.
     */
    old_x_handler = XSetErrorHandler(shm_error_handler);
    XShmAttach(display, shm_info);
    XSync(display, False);
    /* And then reset the old handler here. */
    XSetErrorHandler(old_x_handler);
    
    if (True == shm_error) {
	fprintf(stderr, "darcnes: shm_error == True\n");
	shmdt(shm_info->shmaddr);
	use_shm = False;
	return False;
    }
    
    /* Automatically destroy surface on detach. */
    shmctl(shm_info->shmid, IPC_RMID, NULL);
    
    *buffer = shm_info->shmaddr;
    *image = XShmCreateImage(display, DefaultVisualOfScreen(screen), depth,
			     ZPixmap, shm_info->shmaddr, shm_info, width, height);
    
    return True;
}

void video_leave_deb(void)
{
    deblevel--;
    if (video_active && (!deblevel)) {

    }
}

void video_display_buffer(void)
{
    if (vidxlate) {
	vidxlate(image_buf, buffer_x * buffer_y);
    }

    if (use_shm) {
	XShmPutImage(display, nes_window, nes_gc, nes_image,
		     0, 0, 0, 0, buffer_x, buffer_y, False);
    } else {
	XPutImage(display, nes_window, nes_gc, nes_image, 0, 0, 0, 0, buffer_x, buffer_y);
    }

    XFlush(display);

    if (image_buf == image_buf_0) {
	image_buf = image_buf_1;
	nes_image = nes_image_1;
    } else {
	image_buf = image_buf_0;
	nes_image = nes_image_0;
    }
}

/* joypad emulation */

int ui_register_joypad(struct joypad *pad)
{
    if (ui_joypad) {
	return 0;
    }
    
    ui_joypad = pad;
    return 1;
}

void ui_update_joypad(struct joypad *pad)
{
    /* NOTE: does nothing. may want to do stuff when using real gamepads */
}

void video_events(void)
{
}

char joypad_key_state[12];

struct {
    KeySym keysym;
    int state;
} joypad_keys[] = {
    {XK_s, 8}, {XK_S, 8},
    {XK_a, 9}, {XK_A, 9},
    {XK_bracketleft, 10}, {XK_braceleft, 10},
    {XK_bracketright, 11}, {XK_braceright, 11},
    {XK_Up, 0}, {XK_KP_Up, 4},
    {XK_Down, 1}, {XK_KP_Down, 5},
    {XK_Left, 2}, {XK_KP_Left, 6},
    {XK_Right, 3}, {XK_KP_Right, 7},
};

void handle_joypad_key(int type, KeySym keysym)
{
    int i;

    for (i = (sizeof(joypad_keys) / sizeof(joypad_keys[0])) - 1; i >= 0; i--) {
	if (keysym == joypad_keys[i].keysym) {
	    if (type == KeyPress) {
		joypad_key_state[joypad_keys[i].state] = 1;
	    } else {
		joypad_key_state[joypad_keys[i].state] = 0;
	    }
	}
    }
    
    if (!ui_joypad) {
	return;
    }

    switch (ui_joypad->template->num_buttons) {
	/* NOTE: all cases fall through */
    case 8:
	if (joypad_key_state[11]) {
	    ui_joypad->data |= ui_joypad->template->buttons[7];
	} else {
	    ui_joypad->data &= ~ui_joypad->template->buttons[7];
	}
	
    case 7:
	if (joypad_key_state[10]) {
	    ui_joypad->data |= ui_joypad->template->buttons[6];
	} else {
	    ui_joypad->data &= ~ui_joypad->template->buttons[6];
	}
	
    case 6:
	if (joypad_key_state[9]) {
	    ui_joypad->data |= ui_joypad->template->buttons[5];
	} else {
	    ui_joypad->data &= ~ui_joypad->template->buttons[5];
	}
	
    case 5:
	if (joypad_key_state[8]) {
	    ui_joypad->data |= ui_joypad->template->buttons[4];
	} else {
	    ui_joypad->data &= ~ui_joypad->template->buttons[4];
	}
	
    case 4:
	if (joypad_key_state[7] || joypad_key_state[3]) {
	    ui_joypad->data |= ui_joypad->template->buttons[3];
	} else {
	    ui_joypad->data &= ~ui_joypad->template->buttons[3];
	}
	
    case 3:
	if (joypad_key_state[6] || joypad_key_state[2]) {
	    ui_joypad->data |= ui_joypad->template->buttons[2];
	} else {
	    ui_joypad->data &= ~ui_joypad->template->buttons[2];
	}
	
    case 2:
	if (joypad_key_state[5] || joypad_key_state[1]) {
	    ui_joypad->data |= ui_joypad->template->buttons[1];
	} else {
	    ui_joypad->data &= ~ui_joypad->template->buttons[1];
	}
	
    case 1:
	if (joypad_key_state[4] || joypad_key_state[0]) {
	    ui_joypad->data |= ui_joypad->template->buttons[0];
	} else {
	    ui_joypad->data &= ~ui_joypad->template->buttons[0];
	}
	
    default:
    }
}

/* keypad emulation */

int keypad_register(struct keypad *pad)
{
    if (ui_keypad) {
	return 0;
    }
    
    ui_keypad = pad;
    return 1;
}

void keypad_update(struct keypad *pad)
{
    /* NOTE: does nothing. may want to do stuff when using real keypads */
}

u16 keypad_data;

struct {
    KeySym keysym;
    u16 state;
} keypad_keys[12] = {
    {XK_KP_0, KPD_0},
    {XK_KP_1, KPD_1},
    {XK_KP_2, KPD_2},
    {XK_KP_3, KPD_3},
    {XK_KP_4, KPD_4},
    {XK_KP_5, KPD_5},
    {XK_KP_6, KPD_6},
    {XK_KP_7, KPD_7},
    {XK_KP_8, KPD_8},
    {XK_KP_9, KPD_9},
    {XK_KP_Multiply, KPD_STAR},
    {XK_KP_Add, KPD_HASH},
};

void handle_keypad_key(int type, KeySym keysym)
{
    int i;

    for (i = (sizeof(keypad_keys) / sizeof(keypad_keys[0])) - 1; i >= 0; i--) {
	if (keysym == keypad_keys[i].keysym) {
	    if (type == KeyPress) {
		keypad_data |= keypad_keys[i].state;
	    } else {
		keypad_data &= ~keypad_keys[i].state;
	    }
	}
    }
    
    if (!ui_keypad) {
	return;
    }

    ui_keypad->data = keypad_data;
}

/* key event handler */

void do_key_event(Widget w, XtPointer client_data, XEvent *event, Boolean *continue_dispatch)
{
    KeySym keysym;
    Modifiers dummy;

    XtTranslateKeycode(display, event->xkey.keycode, event->xkey.state, &dummy, &keysym);

    handle_joypad_key(event->type, keysym);
    handle_keypad_key(event->type, keysym);

    if (event->type == KeyPress) {
	switch (keysym) {
	case XK_F1:
	    /* enter debugger */
	    system_flags |= F_BREAK;
	    break;
	    
	case XK_Escape:
	    /* quit */
	    system_flags |= F_QUIT;
	    break;
	}
    }
}

short *xlatepal;

void vidxlate16(char *dest, int size)
{
    int i;

    for (i = size; i > 0; i--) {
	*(((short *)dest)+(i-1)) = xlatepal[ (int) *(dest+(i-1))];
    }
}

unsigned char *vbpfunc8(int line)
{
    return image_buf + (line * buffer_x);
}

unsigned char *vbpfunc16(int line)
{
    return image_buf + (line * buffer_x);
}

/* #include "nespal.c" */

void init_xlate8(int colors, int *red, int *green, int *blue)
{
    int i;
    XColor color;
    
    vid_pre_xlat = malloc(colors);
    vidxlate = NULL;
    video_get_vbp = vbpfunc8;
    
    for (i = 0; i < colors; i++) {
	color.flags = DoRed | DoGreen | DoBlue;
	color.red = red[i];
	color.green = green[i];
	color.blue = blue[i];
	color.red += color.red << 8;
	color.green += color.green << 8;
	color.blue += color.blue << 8;
	if (XAllocColor(display, nes_cmap, &color)) {
	    vid_pre_xlat[i] = color.pixel;
	}
    }
}

void init_xlate16(int colors, int *red, int *green, int *blue)
{
    int i;
    XColor color;

    xlatepal = malloc(colors << 1);
    vid_pre_xlat = malloc(colors);
    vidxlate = (xlatfunc)&vidxlate16;
    video_get_vbp = vbpfunc16;
    
    for (i = 0; i < colors; i++) {
	vid_pre_xlat[i] = i;
	color.flags = DoRed | DoGreen | DoBlue;
	color.red = red[i];
	color.green = green[i];
	color.blue = blue[i];
	color.red += color.red << 8;
	color.green += color.green << 8;
	color.blue += color.blue << 8;
	if (XAllocColor(display, nes_cmap, &color)) {
	    xlatepal[i] = color.pixel;
	}
    }
}

/*
 * $Log: video_x.c,v $
 * Revision 1.18  2000/01/01 03:22:38  nyef
 * added preliminary keypad implementation
 *
 * Revision 1.17  1999/12/31 16:50:30  nyef
 * changed how the joypad was implemented, especially wrt X keyboard events
 *
 * Revision 1.16  1999/06/05 02:41:39  nyef
 * converted to use new joypad interface
 *
 * Revision 1.15  1999/02/15 03:33:54  nyef
 * added patch from Michael Vance for XShm
 *
 * Revision 1.14  1999/02/13 22:23:51  nyef
 * added support for the keypad arrow keys
 *
 * Revision 1.13  1999/02/07 17:55:51  nyef
 * removed double-buffering code
 * added video_setsize() interface
 *
 * Revision 1.12  1999/01/05 04:28:41  nyef
 * hacked up an interface to allow setting the palette.
 *
 * Revision 1.11  1998/08/29 22:09:37  nyef
 * switched to a double-buffered video system. added routines to support
 * the PPU writing directly into the video buffers. eliminated the 8-bit
 * vidxlate routine. changed vidxlate16 to operate using only one buffer.
 * changed the video buffer size to a symbolic constant.
 *
 * Revision 1.10  1998/08/29 14:52:29  nyef
 * added an identity vid_pre_xlat table for use in 16-bitplane modes.
 *
 * Revision 1.9  1998/08/28 03:02:16  nyef
 * added a "vid_pre_xlat" table so that some of the color translation
 * could be done by the PPU. also moved the 8-bit color translation to
 * the PPU and set the 8-bit xlatfinc to be memcpy. (maybe I should
 * double buffer the display?)
 *
 * Revision 1.8  1998/08/15 15:45:34  nyef
 * reworked vidxlate16 to use a normal loop and to copy longs.
 * this should be slightly faster on a pentium.
 *
 * Revision 1.7  1998/08/15 01:08:15  nyef
 * changed unsupported display depth message to indicate current depth.
 *
 * Revision 1.6  1998/08/02 02:03:08  nyef
 * preliminary GUIfication completed.
 *
 * Revision 1.5  1998/08/01 23:12:45  nyef
 * removed the checks for the symbol "GRAPHICS".
 * removed two commented-out printf statements in the input handler.
 *
 * Revision 1.4  1998/08/01 00:57:13  nyef
 * moved old palette values into nespal.c
 *
 * Revision 1.3  1998/07/26 13:22:55  nyef
 * changed vidxlate16 tu use pointer notation and duff's device.
 *
 * Revision 1.2  1998/07/18 18:54:27  nyef
 * changed vidxlate16 to copy shorts instead of chars for a massive speed increase
 *   in 16 bitplane mode.
 *
 * Revision 1.1  1998/07/11 22:19:25  nyef
 * Initial revision
 *
 */
